// ============================================================================ //
// Lightflow Rendering Interface                                                //
// ============================================================================ //
// LfArray.h                                                                    //
// ============================================================================ //
//                                                                              //
// The contents of this file are subject to the Lightflow Rendering Interface   //
// License (the License); you may not use this file except in compliance with   //
// the License. You may obtain a copy of the License by sending an email at     //
// jp@lightflowtech.com with subject LRI-LICENSE.                               //
//                                                                              //
// ============================================================================ //
// Copyright (c) 1995-2000 by Jacopo Pantaleoni. All rights reserved.           //
// ============================================================================ //

#ifndef _LfArray_H
#define _LfArray_H 1

#include <Lightflow/LfArrayBlock.h>


template <class TClass> class LfAPI_CS LfPrimitiveComparator
{
public:
	LfPrimitiveComparator(void) {}
	virtual ~LfPrimitiveComparator(void) {}

public:
	virtual LfInt Compare(const TClass& t1, const TClass& t2) = 0;

	virtual void  OnSwap(LfInt i1, LfInt i2) {}
};

template <class TClass> class LfAPI_CS LfInstanceComparator
{
public:
	LfInstanceComparator(void) {}
	virtual ~LfInstanceComparator(void) {}

public:
	virtual LfInt Compare(const TClass& t1, const TClass& t2) = 0;

	virtual void  OnSwap(LfInt i1, LfInt i2) {}
};

template <class TClass> class LfAPI_CS LfPointerComparator
{
public:
	LfPointerComparator(void) {}
	virtual ~LfPointerComparator(void) {}

public:
	virtual LfInt Compare(TClass t1, TClass t2) = 0;

	virtual void  OnSwap(LfInt i1, LfInt i2) {}
};



template <class TClass> class LfAPI_CS LfPrimitiveArray
{
private:
	LfInt   Size;
	LfInt   MinSize;
	LfInt   RealSize;

	TClass* Table;

public:
	LfPrimitiveArray  (LfInt size = 0, LfInt minsize = 0) : Size(0), MinSize(minsize), RealSize(0), Table(0) { SetSize( size ); }
	LfPrimitiveArray  (const LfPrimitiveArray<TClass>& array ) : Size( 0 ), MinSize( array.MinSize ), RealSize(0), Table(0)
	{
		SetSize( array.Size );

		for (LfInt i=0; i<Size; i++)
			Set( i, array.Get(i) );
	}
	~LfPrimitiveArray (void)
	{
		free( Table );
	}

	inline TClass& operator [] (LfInt i);

	inline void          Add (const TClass& n);
	inline const TClass& Get (LfInt i) const;
	inline void          Set (LfInt i, const TClass& n);

	inline void   SetSize    (LfInt size);
	inline LfInt  GetSize    (void) const { return Size; }
	inline void   Compactify (void);

	inline void   Sort       (LfPrimitiveComparator<TClass>* comparator);

private:
	inline void   Sort       (LfPrimitiveComparator<TClass>* comparator, LfInt p, LfInt r);
	inline LfInt  Partition  (LfPrimitiveComparator<TClass>* comparator, LfInt p, LfInt r);
};


template <class TClass> class LfAPI_CS LfInstanceArray
{
private:
	LfInt   Size;
	LfInt   MinSize;

	LfInstanceBlock<TClass> *Block;

public:
	LfInstanceArray  (LfInt size = 0, LfInt minsize = 0) : Size(0), MinSize(minsize), Block(0) { SetSize( size ); }
	LfInstanceArray  (const LfInstanceArray<TClass>& array ) : Size( 0 ), MinSize( array.MinSize ), Block(0)
	{
		SetSize( array.Size );

		for (LfInt i=0; i<Size; i++)
			Set( i, array.Get(i) );
	}
	~LfInstanceArray (void)
	{
		LfInstanceBlock<TClass> *block, *next;

		block = Block;

		while (block)
		{
			next = block->Next;

			delete block;

			block = next;
		}
	}

	inline TClass& operator [] (LfInt i);

	inline void          Add (const TClass& n);
	inline const TClass& Get (LfInt i) const;
	inline void          Set (LfInt i, const TClass& n);

	inline void    SetSize    (LfInt size);
	inline LfInt   GetSize    (void) const { return Size; }
	inline void    Compactify (void);

	inline void    Sort       (LfInstanceComparator<TClass>* comparator);

private:
	inline void    Sort       (LfInstanceComparator<TClass>* comparator, LfInt p, LfInt r);
	inline LfInt   Partition  (LfInstanceComparator<TClass>* comparator, LfInt p, LfInt r);
};


template <class TClass> class LfAPI_CS LfPointerArray
{
private:
	LfInt   Size;
	LfInt   MinSize;
	LfInt   RealSize;

	TClass* Table;

public:
	LfPointerArray  (LfInt size = 0, LfInt minsize = 0) : Size(0), MinSize(minsize), RealSize(0), Table(0) { SetSize( size ); }
	LfPointerArray  (const LfPointerArray<TClass>& array ) : Size( 0 ), MinSize( array.MinSize ), RealSize(0), Table(0)
	{
		SetSize( array.Size );

		for (LfInt i=0; i<Size; i++)
			Set( i, array.Get(i) );
	}
	~LfPointerArray (void)
	{
		free( Table );
	}

	inline TClass operator [] (LfInt i) const;

	inline void   Add (TClass n);
	inline TClass Get (LfInt i) const;
	inline void   Set (LfInt i, TClass n);

	inline void   SetSize    (LfInt size);
	inline LfInt  GetSize    (void) const { return Size; }
	inline void   Compactify (void);

	inline void   Sort       (LfPointerComparator<TClass>* comparator);

private:
	inline void   Sort       (LfPointerComparator<TClass>* comparator, LfInt p, LfInt r);
	inline LfInt  Partition  (LfPointerComparator<TClass>* comparator, LfInt p, LfInt r);
};

// ---------------------------------------- INSTANTIATION ----------------------------------------- //
LfTEMPLATE_CS( LfPrimitiveArray<LfByte> );
LfTEMPLATE_CS( LfPrimitiveArray<LfWord> );
LfTEMPLATE_CS( LfPrimitiveArray<LfInt> );
LfTEMPLATE_CS( LfPrimitiveArray<LfFloat> );
LfTEMPLATE_CS( LfPrimitiveArray<LfVector2> );
LfTEMPLATE_CS( LfPrimitiveArray<LfVector3> );
LfTEMPLATE_CS( LfPrimitiveArray<LfVector4> );
LfTEMPLATE_CS( LfPrimitiveArray<LfMatrix2> );
LfTEMPLATE_CS( LfPrimitiveArray<LfMatrix3> );
LfTEMPLATE_CS( LfPrimitiveArray<LfMatrix4> );
LfTEMPLATE_CS( LfPrimitiveArray<LfTransform> );

LfTEMPLATE_CS( LfInstanceArray<LfByte> );
LfTEMPLATE_CS( LfInstanceArray<LfWord> );
LfTEMPLATE_CS( LfInstanceArray<LfInt> );
LfTEMPLATE_CS( LfInstanceArray<LfFloat> );
LfTEMPLATE_CS( LfInstanceArray<LfVector2> );
LfTEMPLATE_CS( LfInstanceArray<LfVector3> );
LfTEMPLATE_CS( LfInstanceArray<LfVector4> );
LfTEMPLATE_CS( LfInstanceArray<LfMatrix2> );
LfTEMPLATE_CS( LfInstanceArray<LfMatrix3> );
LfTEMPLATE_CS( LfInstanceArray<LfMatrix4> );
LfTEMPLATE_CS( LfInstanceArray<LfTransform> );
// ---------------------------------------- IMPLEMENTATION ---------------------------------------- //


#define USE_SEDGEWICK_QUICKSORT 0


template <class TClass> TClass& LfPrimitiveArray<TClass>::operator[](LfInt i)
{
#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfPrimitiveArray::operator[]", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfPrimitiveArray::operator[]", LfFault) );
		return Table[0];
	}
#endif
	return Table[ i ];
}
template <class TClass> void LfPrimitiveArray<TClass>::Add(const TClass& n)
{
	SetSize( Size+1 );

	Set( Size-1, n );
}
template <class TClass> const TClass& LfPrimitiveArray<TClass>::Get(LfInt i) const
{
#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfPrimitiveArray::Get()", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfPrimitiveArray::Get()", LfFault) );
		return Table[0];
	}
#endif
	return Table[ i ];
}
template <class TClass> void LfPrimitiveArray<TClass>::Set(LfInt i, const TClass& n)
{
#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfPrimitiveArray::Set()", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfPrimitiveArray::Set()", LfFault) );
		return;
	}
#endif
	Table[ i ] = n;
}
template <class TClass> void LfPrimitiveArray<TClass>::SetSize(LfInt size)
{
	if (! size)
	{
		free( Table ); Table = 0;

		Size = 0;
		RealSize = 0;
	}
	else if (size < Size)
	{
		Size = size;

		// Search if there is superfluous space at the end
		if ((Size*2 <= RealSize) && (RealSize >= MinSize*2))
		{
			RealSize = Size;

			Table = (TClass*)realloc( Table, RealSize*sizeof( TClass ) );
		}
	}
	else if (size > Size)
	{
		Size = size;

		// Search if there is empty space at the end
		if (Size <= RealSize)
			return;

		// No empty space: expand the block
		if (! RealSize)
		{
			if (Size < MinSize)
				RealSize = MinSize;
			else
				RealSize = Size;
		}
		else
		{
			while (RealSize < Size)
				RealSize *= 2;
		}

		Table = (TClass*)realloc( Table, RealSize*sizeof( TClass ) );
	}
}

template <class TClass> void LfPrimitiveArray<TClass>::Compactify(void)
{
	if (! Size)
	{
		free( Table ); Table = 0;

		RealSize = 0;
	}
	else if (Size != RealSize)
	{
		if (Size < MinSize)
			RealSize = MinSize;
		else
			RealSize = Size;

		Table = (TClass*)realloc( Table, RealSize*sizeof( TClass ) );
	}
}
template <class TClass> void LfPrimitiveArray<TClass>::Sort(LfPrimitiveComparator<TClass>* cmp)
{
	Sort( cmp, 0, Size-1 );
}

template <class TClass> void LfPrimitiveArray<TClass>::Sort(LfPrimitiveComparator<TClass>* cmp, LfInt p, LfInt r)
{
#if USE_SEDGEWICK_QUICKSORT
	LfInt  i;
	LfInt  a, b;
	TClass ae, be, tmp;

	a = (p+r)/2;
	b = r-1;

	cmp->OnSwap( a, b );

	tmp = Get(a);
	Set( a, Get(b) );
	Set( b, tmp );

	a = p;
	ae = Get(a);
	be = Get(b);

	if (cmp->Compare( ae, be ) < 0)
	{
		cmp->OnSwap( a, b );

		Set( a, be );
		Set( b, ae );
	}

	b = r;
	ae = Get(a);
	be = Get(b);

	if (cmp->Compare( ae, be ) < 0)
	{
		cmp->OnSwap( a, b );

		Set( a, be );
		Set( b, ae );
	}

	a = r-1;
	ae = Get(a);
	be = Get(b);

	if (cmp->Compare( ae, be ) < 0)
	{
		cmp->OnSwap( a, b );

		Set( a, be );
		Set( b, ae );
	}

	i = Partition( cmp, p+1, r-1 );

	Sort( cmp, p, i-1 );
	Sort( cmp, i+1, r );
#else
	LfInt q;
	while (p < r)
	{
		q = Partition( cmp, p, r );
		Sort( cmp, p, q );
		p = q + 1;
	}
#endif
}
template <class TClass> LfInt LfPrimitiveArray<TClass>::Partition(LfPrimitiveComparator<TClass>* cmp, LfInt p, LfInt r)
{
#if USE_SEDGEWICK_QUICKSORT
	TClass  x, tmp;
	LfInt   i, j;

	i = p-1;
	j = r;

	while (1)
	{
		while (cmp->Compare( Get(++i), x ) <= 0);
		while (cmp->Compare( x, Get(--j) ) <= 0) if (j == p) break;

		if (i >= j) break;

		cmp->OnSwap( i, j );

		tmp = Get(i);
		Set( i, Get(j) );
		Set( j, tmp );
	}
	return i;
#else
	TClass  x, tmp;
	LfInt   i, j;

	x = Get(p);

	i = p - 1;
	j = r + 1;

	while (1)
	{
		while(1)
		{
			if (cmp->Compare( Get(--j), x ) <= 0)
				break;

			if (j <= 0)
				break;
		}

		while(1)
		{
			if (cmp->Compare( Get(++i), x ) >= 0)
				break;

			if (i >= Size-1)
				break;
		}

		if (i < j)
		{
			cmp->OnSwap( i, j );

			tmp = Get(i);
			Set( i, Get(j) );
			Set( j, tmp );
		}
		else return j;
	}
	return 0;
#endif
}




template <class TClass> TClass& LfInstanceArray<TClass>::operator[](LfInt i)
{
	LfInstanceBlock<TClass>* block = Block;
	LfInt size = 0;

#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfInstanceArray::operator[]", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfInstanceArray::operator[]", LfFault) );
		return block->Table[0];
	}
#endif
	while (block)
	{
		if (i < size + block->Size)
			return block->Table[ i - size ];

		size += block->Size;
		block = block->Next;
	}
	return Block->Table[0];
}
template <class TClass> void LfInstanceArray<TClass>::Add(const TClass& n)
{
	SetSize( Size+1 );

	Set( Size-1, n );
}
template <class TClass> const TClass& LfInstanceArray<TClass>::Get(LfInt i) const
{
	LfInstanceBlock<TClass>* block = Block;
	LfInt size = 0;

#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfInstanceArray::Get()", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfInstanceArray::Get()", LfFault) );
		return Block->Table[0];
	}
#endif
	while (block)
	{
		if (i < size + block->Size)
			return block->Table[ i - size ];

		size += block->Size;
		block = block->Next;
	}
	return Block->Table[0];
}
template <class TClass> void LfInstanceArray<TClass>::Set(LfInt i, const TClass& n)
{
	LfInstanceBlock<TClass>* block = Block;
	LfInt size = 0;

#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfInstanceArray::Set()", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfInstanceArray::Set()", LfFault) );
		return;
	}
#endif
	while (block)
	{
		if (i < size + block->Size)
		{
			block->Table[ i - size ] = n;
			return;
		}

		size += block->Size;
		block = block->Next;
	}
}
template <class TClass> void LfInstanceArray<TClass>::SetSize(LfInt size)
{
	if (! size)
	{
		LfInstanceBlock<TClass> *block, *next;

		Size = 0;

		block = Block;

		while (block)
		{
			next = block->Next;

			delete block;

			block = next;
		}

		Block = 0;
	}
	else if (size < Size)
	{
		// Search if there are superfluous blocks at the end
		LfInstanceBlock<TClass> *block, *next;
		LfInt                   cursize = 0;

		Size = size;

		block = Block;
		while (block)
		{
			cursize += block->Size;

			if ((cursize >= size) && (cursize >= MinSize))
			{
				// Start deleting blocks
				next = block->Next;

				block->Next = 0;

				block = next;

				while (block)
				{
					next = block->Next;

					delete block;

					block = next;
				}
				return;
			}

			block = block->Next;
		}
	}
	else if (size > Size)
	{
		LfInstanceBlock<TClass> *block, *prev;
		LfInt                   cursize = 0;

		Size = size;

		// Search if there is empty space at the end
		prev  = 0;
		block = Block;
		while (block)
		{
			prev = block;

			cursize += block->Size;

			if (cursize >= size)
				return;

			block = block->Next;
		}

		// No empty space: allocate a new block and add it to the queue
		if ((! Block) && (size < MinSize))
			block = new LfInstanceBlock<TClass>( MinSize );
		else
			block = new LfInstanceBlock<TClass>( size );

		if (prev)
			prev->Next = block;
		else
			Block = block;
	}
}
template <class TClass> void LfInstanceArray<TClass>::Compactify(void)
{
	LfInstanceBlock<TClass> *block, *next, *newblock;
	LfInt                    size = 0;
	LfInt                    index;

	if (Size <= MinSize)
		return;

	newblock = new LfInstanceBlock<TClass>( Size );

	block = Block;
	while (block)
	{
		index = block->Size;
		if (index > Size - size)
			index = Size - size;

		for (LfInt i = 0; i < index; i++)
			newblock->Table[ size + i ] = block->Table[ i ];

		size += block->Size;

		next = block->Next;

		delete block;

		block = next;
	}

	Block = newblock;
}
template <class TClass> void LfInstanceArray<TClass>::Sort(LfInstanceComparator<TClass>* cmp)
{
/*	LfInt n = Size / 2;
	for (LfInt i = 0; i < n; i++)
	{
		LfInt j = rand() % n + n;

		cmp->OnSwap( i, j );

		TClass tmp = Get(i);
		Set( i, Get(j) );
		Set( j, tmp );
	}*/

	Sort( cmp, 0, Size-1 );
}
template <class TClass> void LfInstanceArray<TClass>::Sort(LfInstanceComparator<TClass>* cmp, LfInt p, LfInt r)
{
#if USE_SEDGEWICK_QUICKSORT
	LfInt  i;
	LfInt  a, b;
	TClass tmp;

	a = (p+r)/2;
	b = r-1;

	cmp->OnSwap( a, b );

	tmp = Get(a);
	Set( a, Get(b) );
	Set( b, tmp );

	a = p;

	if (cmp->Compare( Get(a), Get(b) ) < 0)
	{
		cmp->OnSwap( a, b );

		tmp = Get(a);
		Set( a, Get(b) );
		Set( b, tmp );
	}

	b = r;

	if (cmp->Compare( Get(a), Get(b) ) < 0)
	{
		cmp->OnSwap( a, b );

		tmp = Get(a);
		Set( a, Get(b) );
		Set( b, tmp );
	}

	a = r-1;

	if (cmp->Compare( Get(a), Get(b) ) < 0)
	{
		cmp->OnSwap( a, b );

		tmp = Get(a);
		Set( a, Get(b) );
		Set( b, tmp );
	}

	i = Partition( cmp, p+1, r-1 );

	Sort( cmp, p, i-1 );
	Sort( cmp, i+1, r );
#else
	LfInt q;
	while (p < r)
	{
		q = Partition( cmp, p, r );
		Sort( cmp, p, q );
		p = q + 1;
	}
#endif
}
template <class TClass> LfInt LfInstanceArray<TClass>::Partition(LfInstanceComparator<TClass>* cmp, LfInt p, LfInt r)
{
#if USE_SEDGEWICK_QUICKSORT
	TClass  x, tmp;
	LfInt   i, j;

	i = p-1;
	j = r;

	while (1)
	{
		while (cmp->Compare( Get(++i), x ) <= 0);
		while (cmp->Compare( x, Get(--j) ) <= 0) if (j == p) break;

		if (i >= j) break;

		cmp->OnSwap( i, j );

		tmp = Get(i);
		Set( i, Get(j) );
		Set( j, tmp );
	}
	return i;
#else
	TClass  x, tmp;
	LfInt   i, j;

	x = Get(p);

	i = p - 1;
	j = r + 1;

	while (1)
	{
		while(1)
		{
			if (cmp->Compare( Get(--j), x ) <= 0)
				break;

			if (j <= 0)
				break;
		}

		while(1)
		{
			if (cmp->Compare( Get(++i), x ) >= 0)
				break;

			if (i >= Size-1)
				break;
		}

		if (i < j)
		{
			cmp->OnSwap( i, j );

			tmp = Get(i);
			Set( i, Get(j) );
			Set( j, tmp );
		}
		else return j;
	}
	return 0;
#endif
}





template <class TClass> TClass LfPointerArray<TClass>::operator[](LfInt i) const
{
#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfPointerArray::operator[]", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfPointerArray::operator[]", LfFault) );
		return Table[0];
	}
#endif
	return Table[ i ];
}
template <class TClass> void LfPointerArray<TClass>::Add(TClass n)
{
	SetSize( Size+1 );

	Set( Size-1, n );
}
template <class TClass> TClass LfPointerArray<TClass>::Get(LfInt i) const
{
#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfPointerArray::Get()", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfPointerArray::Get()", LfFault) );
		return Table[0];
	}
#endif
	return Table[ i ];
}
template <class TClass> void LfPointerArray<TClass>::Set(LfInt i, TClass n)
{
#if LfASSERTS
	if (i<0 || i>=Size)
	{
//		throw LfError("bad index", "LfPointerArray::Set()", LfFault);
		LfDefaultErrorHandler( LfError("bad index", "LfPointerArray::Set()", LfFault) );
		return;
	}
#endif
	Table[ i ] = n;
}
template <class TClass> void LfPointerArray<TClass>::SetSize(LfInt size)
{
	if (! size)
	{
		free( Table ); Table = 0;

		Size = 0;
		RealSize = 0;
	}
	else if (size < Size)
	{
		Size = size;

		// Search if there is superfluous space at the end
		if ((Size*2 <= RealSize) && (RealSize >= MinSize*2))
		{
			RealSize = Size;

			Table = (TClass*)realloc( Table, RealSize*sizeof( TClass ) );
		}
	}
	else if (size > Size)
	{
		Size = size;

		// Search if there is empty space at the end
		if (Size <= RealSize)
			return;

		// No empty space: expand the block
		if (! RealSize)
		{
			if (Size < MinSize)
				RealSize = MinSize;
			else
				RealSize = Size;
		}
		else
		{
			while (RealSize < Size)
				RealSize *= 2;
		}

		Table = (TClass*)realloc( Table, RealSize*sizeof( TClass ) );
	}
}

template <class TClass> void LfPointerArray<TClass>::Compactify(void)
{
	if (! Size)
	{
		free( Table ); Table = 0;

		RealSize = 0;
	}
	else if (Size != RealSize)
	{
		if (Size < MinSize)
			RealSize = MinSize;
		else
			RealSize = Size;

		Table = (TClass*)realloc( Table, RealSize*sizeof( TClass ) );
	}
}
template <class TClass> void LfPointerArray<TClass>::Sort(LfPointerComparator<TClass>* cmp)
{
/*	LfInt n = Size / 2;
	for (LfInt i = 0; i < n; i++)
	{
		LfInt j = rand() % n + n;

		cmp->OnSwap( i, j );

		TClass tmp = Get(i);
		Set( i, Get(j) );
		Set( j, tmp );
	}*/

	Sort( cmp, 0, Size-1 );
}

template <class TClass> void LfPointerArray<TClass>::Sort(LfPointerComparator<TClass>* cmp, LfInt p, LfInt r)
{
#if USE_SEDGEWICK_QUICKSORT
	LfInt  i;
	LfInt  a, b;
	TClass tmp;

	a = (p+r)/2;
	b = r-1;

	cmp->OnSwap( a, b );

	tmp = Get(a);
	Set( a, Get(b) );
	Set( b, tmp );

	a = p;

	if (cmp->Compare( Get(a), Get(b) ) < 0)
	{
		cmp->OnSwap( a, b );

		tmp = Get(a);
		Set( a, Get(b) );
		Set( b, tmp );
	}

	b = r;

	if (cmp->Compare( Get(a), Get(b) ) < 0)
	{
		cmp->OnSwap( a, b );

		tmp = Get(a);
		Set( a, Get(b) );
		Set( b, tmp );
	}

	a = r-1;

	if (cmp->Compare( Get(a), Get(b) ) < 0)
	{
		cmp->OnSwap( a, b );

		tmp = Get(a);
		Set( a, Get(b) );
		Set( b, tmp );
	}

	i = Partition( cmp, p+1, r-1 );

	Sort( cmp, p, i-1 );
	Sort( cmp, i+1, r );
#else
	LfInt q;
	while (p < r)
	{
		q = Partition( cmp, p, r );
		Sort( cmp, p, q );
		p = q + 1;
	}
#endif
}
template <class TClass> LfInt LfPointerArray<TClass>::Partition(LfPointerComparator<TClass>* cmp, LfInt p, LfInt r)
{
#if USE_SEDGEWICK_QUICKSORT
	TClass  x, tmp;
	LfInt   i, j;

	i = p-1;
	j = r;

	while (1)
	{
		while (cmp->Compare( Get(++i), x ) <= 0);
		while (cmp->Compare( x, Get(--j) ) <= 0) if (j == p) break;

		if (i >= j) break;

		cmp->OnSwap( i, j );

		tmp = Get(i);
		Set( i, Get(j) );
		Set( j, tmp );
	}
	return i;
#else
	TClass  x, tmp;
	LfInt   i, j;

	x = Get(p);

	i = p - 1;
	j = r + 1;

	while (1)
	{
		while(1)
		{
			if (cmp->Compare( Get(--j), x ) <= 0)
				break;

			if (j <= 0)
				break;
		}

		while(1)
		{
			if (cmp->Compare( Get(++i), x ) >= 0)
				break;

			if (i >= Size-1)
				break;
		}

		if (i < j)
		{
			cmp->OnSwap( i, j );

			tmp = Get(i);
			Set( i, Get(j) );
			Set( j, tmp );
		}
		else return j;
	}
	return 0;
#endif
}


#endif
